var builtin = builtin || {};
var _LINK_UI_SDK = _LINK_UI_SDK || {};

builtin.android = {
  createChannel: (prefix, callback) => {
    let channel = prefix + Math.random().toString(36).substring(2, 15);
    builtin.android.state.channelMap[channel] = callback;
    return channel;
  },
  handleCallback: (channel, message, error) => {
    let parsedMessage = new AndroidCallbackMessage(
      channel,
      message,
      error ? new Error(error) : undefined
    );
    if (builtin.android.state.channelMap[channel]) {
      builtin.android.state.channelMap[channel](parsedMessage);
    } else {
      console.error("No callback found for channel", channel);
    }
  },

  /**
   * Check if the Android SDK is available and injected into the page
   * @returns True if the Android SDK is available
   */
  isAndroidSDK: () => window.AndroidBluetooth !== undefined,

  state: {
    channelMap: {
      onGATTServerDisconnected: (msg) => {
        if (msg.error) {
          console.error(msg.error);
          return;
        }

        let parsedMessage = JSON.parse(msg.message);
        let device = builtin.android.state.deviceMap[parsedMessage.address];
        if (device) {
          device.gatt.connected = false;
          device.dispatchEvent(new Event("gattserverdisconnected"));
        } else {
          console.error("Device not found for disconnect", parsedMessage.address);
        }
      },
      onCharacteristicChanged: (msg) => {
        if (msg.error) {
          console.error(msg.error);
          return;
        }

        let parsedMessage = JSON.parse(msg.message);
        let characteristic = builtin.android.state.notificationMap[parsedMessage.characteristicUUID];
        if (characteristic) {
          characteristic.value = Uint8Array.from(
            (parsedMessage.hex.match(/.{1,2}/g) ?? []).map((byte) =>
              parseInt(byte, 16)
            )
          );
          characteristic.dispatchEvent(new Event("characteristicvaluechanged"));
        } else {
          console.error("Characteristic not found for notification", parsedMessage.characteristicUUID);
        }
      }
    },
    deviceMap: {},
    notificationMap: {},
  },
};

/**
 * Message Format for Android Callbacks
 */
class AndroidCallbackMessage {
  constructor(channel, message, error) {
    this.channel = channel;
    this.message = message;
    this.error = error;
  }

  static fromObject(object) {
    if (typeof object !== "object") {
      throw new Error("Expected object");
    }

    if (typeof object.channel !== "string") {
      throw new Error("Expected string for channel");
    }

    if (object.message !== undefined && typeof object.message !== "string") {
      throw new Error("Expected string for message");
    }

    if (object.error !== undefined && typeof object.error !== "string") {
      throw new Error("Expected string for error");
    }

    return new AndroidCallbackMessage(
      object.channel,
      object.message,
      object.error
    );
  }
}

/**
 * Polyfill for BluetoothDevice
 */
class AndroidBluetoothDevice extends EventTarget {
  constructor(id, name) {
    super();
    this.id = id;
    this.name = name;
    this.gatt = new AndroidBluetoothRemoteGATTServer(this);
    builtin.android.state.deviceMap[id] = this;
  }

  forget() {
    console.warn("forget() is not implemented for AndroidBluetoothDevice");
  }

  /**
   * Android already tells us the advertised services, so we can just return success (undefined)
   */
  watchAdvertisements() {
    return new Promise((resolve) => resolve(undefined));
  }

  static fromObject(object) {
    if (typeof object !== "object") {
      throw new Error("Expected object");
    }
    if (typeof object.address !== "string") {
      throw new Error("Expected string for id");
    }
    if (typeof object.name !== "string") {
      throw new Error("Expected string for name");
    }
    return new AndroidBluetoothDevice(
      object.address,
      object.name,
    );
  }
}

/**
 * Polyfill for BluetoothRemoteGATTServer
 */
class AndroidBluetoothRemoteGATTServer {
  constructor(device) {
    this.connected = false;
    this.device = device;
  }

  async connect() {
    return new Promise((resolve, reject) => {
      const channel = builtin.android.createChannel("connect", (msg) => {
        if (msg.error) {
          reject(new Error(msg.error));
          return;
        }
        this.connected = true;
        resolve(this);
      });

      window.AndroidBluetooth.gattConnect(channel, JSON.stringify({address: this.device.id}));
      setTimeout(() => reject("gattConnect timeout"), 60 * 1000);
    });
  }

  async disconnect() {
    return new Promise((resolve, reject) => {
      const channel = builtin.android.createChannel("disconnect", (msg) => {
        if (msg.error) {
          reject(new Error(msg.error));
          return;
        }
        this.connected = false;
        resolve(this);
      });
      window.AndroidBluetooth.gattDisconnect(channel, JSON.stringify({address: this.device.id}));
      setTimeout(() => reject("gattDisconnect timeout"), 60 * 1000);
    });
  }

  async getPrimaryService(uuid) {
    return new Promise((resolve, reject) => {
      const channel = builtin.android.createChannel("getPrimaryService", (msg) => {
        if (msg.error) {
          reject(new Error(msg.error));
          return;
        }

        resolve(
          new AndroidBluetoothRemoteGATTService(this, uuid)
        );
      });
      window.AndroidBluetooth.getPrimaryService(
        channel,
        JSON.stringify({ address: this.device.id, serviceUUID: uuid })
      );
      setTimeout(() => reject("getPrimaryService timeout"), 60 * 1000);
    });
  }

  async getPrimaryServices(uuid) {
    return [await this.getPrimaryService(uuid)];
  }

  
}

/**
 * Polyfill for BluetoothRemoteGATTService
 */
class AndroidBluetoothRemoteGATTService {
  constructor(server, uuid) {
    this.server = server;
    this.uuid = uuid;
    this.device = server.device;
    this.isPrimary = true;
  }

  async getCharacteristic(uuid) {
    return new Promise((resolve, reject) => {
      const channel = builtin.android.createChannel("getCharacteristic", (msg) => {
        if (msg.error) {
          reject(new Error(msg.error));
          return;
        }

        resolve(
          new AndroidBluetoothRemoteGATTCharacteristic(this, uuid)
        );
      });
      window.AndroidBluetooth.getCharacteristic(
        channel,
        JSON.stringify({ address: this.device.id, serviceUUID: this.uuid, characteristicUUID: uuid })
      );
      setTimeout(() => reject("getCharacteristic timeout"), 60 * 1000);
    });
  }

  async getCharacteristics(uuid) {
    return [await this.getCharacteristic(uuid)]
  }
}

/**
 * Polyfill for BluetoothRemoteGATTCharacteristic
 */
class AndroidBluetoothRemoteGATTCharacteristic extends EventTarget {
  constructor(service, uuid) {
    super();
    this.service = service;
    this.uuid = uuid;
    this.properties = {}; // Unsupported
    this.value = undefined;
  }

  getDescriptor() {
    return new Promise((_, reject) =>
      reject(
        new Error(
          "getDescriptor() is not implemented on AndroidBluetoothRemoteGATTCharacteristic"
        )
      )
    );
  }

  getDescriptors() {
    return new Promise((_, reject) =>
      reject(
        new Error(
          "getDescriptors() is not implemented on AndroidBluetoothRemoteGATTCharacteristic"
        )
      )
    );
  }

  readValue() {
    return new Promise((resolve, reject) => {
      delete builtin.android.state.channelMap[this.notificationChannel];
      this.notificationChannel = undefined;

      const channel = builtin.android.createChannel(
        "readValue",
        (msg) => {
          if (msg.error) {
            reject(new Error(msg.error));
            return;
          }
          
          this.value = Uint8Array.from(
            (msg.message.hex.match(/.{1,2}/g) ?? []).map((byte) =>
              parseInt(byte, 16)
            )
          );
          resolve(
            this.value
          );
        }
      );
      window.AndroidBluetooth.readCharacteristic(
        channel,
        JSON.stringify({ address: this.service.device.id, serviceUUID: this.service.uuid, characteristicUUID: this.uuid })
      );
    });
  }

  writeValueWithResponse(value) {
    return new Promise((resolve, reject) => {
      const channel = builtin.android.createChannel(
        "writeValueWithResponse",
        (msg) => (msg.error ? reject(new Error(msg.error)) : resolve())
      );
      window.AndroidBluetooth.writeCharacteristic(
        channel,
        JSON.stringify({ address: this.service.device.id, serviceUUID: this.service.uuid, characteristicUUID: this.uuid, hex: [...value].map((x) => x.toString(16).padStart(2, "0")).join("") })
      );
    });
  }

  writeValueWithoutResponse(value) {
    return this.writeValueWithResponse(value);
  }

  startNotifications() {
    return new Promise((resolve, reject) => {
      builtin.android.state.notificationMap[this.uuid] = this;
      const channel = builtin.android.createChannel(
        "startNotifications",
        (msg) => (msg.error ? reject(new Error(msg.error)) : resolve())
      );
      window.AndroidBluetooth.startNotifications(
        channel,
        JSON.stringify({ address: this.service.device.id, serviceUUID: this.service.uuid, characteristicUUID: this.uuid })
      );
    });
  }

  stopNotifications() {
    return new Promise((resolve, reject) => {
      delete builtin.android.state.notificationMap[this.uuid];

      const channel = builtin.android.createChannel(
        "stopNotifications",
        (msg) => (msg.error ? reject(new Error(msg.error)) : resolve())
      );
      window.AndroidBluetooth.stopNotifications(
        channel,
        JSON.stringify({ address: this.service.device.id, serviceUUID: this.service.uuid, characteristicUUID: this.uuid })
      );
    });
  }
}

function createProperty(value) {
  var _value = value;

  function _get() {
      return _value;
  }

  function _set(v) {
      _value = v;
  }

  return {
      "get": _get,
      "set": _set
  };
};

function makePropertyWritable(objBase, objScopeName, propName, initValue) {
  var newProp,
      initObj;

  if (objBase && objScopeName in objBase && propName in objBase[objScopeName]) {
      if(typeof initValue === "undefined") {
          initValue = objBase[objScopeName][propName];
      }

      newProp = createProperty(initValue);

      try {
          Object.defineProperty(objBase[objScopeName], propName, newProp);
      } catch (e) {
          initObj = {};
          initObj[propName] = newProp;
          try {
              objBase[objScopeName] = Object.create(objBase[objScopeName], initObj);
          } catch(e) {}
      }
  }
};

makePropertyWritable(window, "navigator", "bluetooth");
window.SDKCallbackHandler = builtin.android.handleCallback;
_LINK_UI_SDK.oauth = function(authorizeUrl, interceptUrl, returnUrl, config) {
    AndroidOAuth.oauth(authorizeUrl, interceptUrl, returnUrl, JSON.stringify(config));
}
navigator.bluetooth = {
  /**
   * Maps the availability of bluetooth to the availability of the Android SDK
   */
  getAvailability: () =>
    new Promise((resolve) => resolve(builtin.android.isAndroidSDK())),

  /**
   * Map the requestDevice function to the Android SDK, map response of the Android SDK to a polyfill BluetoothDevice
   */
  requestDevice: async (options) => {
    return new Promise((resolve, reject) => {
      const channel = builtin.android.createChannel("requestDevice", (msg) => {
        delete builtin.android.state.channelMap[channel];
        if (msg.error) {
          reject(new Error(msg.error));
        } else {
          resolve(
            AndroidBluetoothDevice.fromObject(JSON.parse(msg.message))
          );
        }
      });
      window.AndroidBluetooth.requestDevice(
        channel,
        JSON.stringify(options)
      );
      setTimeout(() => reject("requestDevice timeout"), 60 * 1000);
    });
  },
};